#include <fstream>
#include <vector>
/**
 * HSet represents a set on the section as
 *   X = c + B*r
 * where
 * c - is the origin of the coordinate system
 * B - is a linear coordinate system
 * r - is a box centered at the origin
 * We assume that X[0] = 0 and this is the 'section variable'.
 * The matrix inv_B is a rigorous inverse of B.
 * 
 * The set is embeded into Poincare section by an orthogonal mapping E, i.e.
 *   Y = c + E*B*r
 * The matrix E maps first coordinate of B*r into ortogonal direction to the section.
 * The remaining columns of E form an orthogonal basis on the section.
 * 
 * The matrix inv_E is a rigorous inverse of E.
 *  
 */ 
struct HSet{
  HSet(int m) 
    : c(m), B(m,m), inv_B(m,m), r(m), 
      E(IMatrix::Identity(m)), inv_E(IMatrix::Identity(m)),
      section(IVector(m),IVector(m))
  {}

  void setCenterAndSection(LDVector _c, LDVector _n){
    this->c = vectalg::convertObject<IVector>(_c);
    IVector n = vectalg::convertObject<IVector>(_n);
    this->E.column(0) = n;
    orthonormalizeColumns(this->E);
    this->inv_E = Transpose(this->E);
    this->section = IAffineSection(this->c,n);
  }

  void setCenter(LDVector _c){
    this->c = vectalg::convertObject<IVector>(_c);
    this->section.setOrigin(this->c);
  }

  void setCoordinateSystem(IMatrix _B){
    this->B = _B;
    this->inv_B = matrixAlgorithms::krawczykInverse(B);
  }

  void setCoordinateSystem(LDMatrix _B){
    setCoordinateSystem(vectalg::convertObject<IMatrix>(_B));
  }

  IVector c;
  IMatrix B, inv_B;
  IVector r;
  IMatrix E, inv_E;
  
  IAffineSection section;
  int iterations;
};

// #####################################################################

struct ProofData{
  LDVector p1, p2;
  LDMatrix C1, C2;

  HSet* fp1;
  HSet* fp2;
  std::vector<HSet*> h12;
  std::vector<HSet*> h21;

 /// This function inits all the data for the proof. It is called from the main.cpp
  void init(LDKSPoincareMap& pm, int dim, double setSize){
    /// first we read periodic points from file (precomputed)
    this->readPeriodicPoints(dim);
    /// and coordinate systems built on approximate eigenvectors (precomputed)
    this->readCoordinateSystems(dim);

    /// given this date we construct h-sets at periodic points
    this->fp1 = initHSetAtPeriodicPoint(pm,p1,C1,2,setSize);
    this->fp2 = initHSetAtPeriodicPoint(pm,p2,C2,4,setSize);

    /// read and construct heteroclinic chains
    this->initP1P2Chain(pm,dim);
    this->initP2P1Chain(pm,dim);
  }

  /// This routine reads precomputed periodic point from a file.
  /// Then, the point is resized to the dimension in which actual computation are performed.
  /// Note, we did experiments with varying dimensions, therefore the routine is a bit more general than actually needed for the final proof.
  void readPeriodicPoint(const char* filename, int dim, LDVector& p){
    std::ifstream in;
    in.open(filename);
    int D = capd::min(20,dim);
    do{
      in >> p;
    }while(p.dimension()!=D);
    in.close();
    if(D!=dim)
      p = resizeVector(p,dim);
  }

  /// Read two approximate periodic points from the files.
  void readPeriodicPoints(int dim){
    readPeriodicPoint("data/perpoint2.dat",dim,p1);
    readPeriodicPoint("data/perpoint4.dat",dim,p2);
  }

  /// Read coordinate system at periodic point from a file (precomputed).
  void readCoordinateSystem(const char* filename, int dim, LDMatrix& C){
    std::ifstream in;
    in.open(filename);
    do{
      in >> C;
    }while(C.numberOfRows()!=dim);
    orthonormalizeColumns(C);
    in.close();
  }

  /// Read coorrdinate systems at fixed point and period two point.
  void readCoordinateSystems(int dim){
    int D = capd::min(dim,14);
    readCoordinateSystem("data/cs2.dat",D,C1);
    readCoordinateSystem("data/cs4.dat",D,C2);
    C1 = resizeMatrix(C1,dim);
    C2 = resizeMatrix(C2,dim);
  }
  
  /// This routine creates an HSet at periodic point. 
  HSet* initHSetAtPeriodicPoint(LDKSPoincareMap& pm, const LDVector& p, const LDMatrix& B, int period, double r){
    int dim = p.dimension();
    HSet* hset = new HSet(dim); 
    LDVector n = pm.setOrthogonalSectionAt(p);       
    hset->setCenterAndSection(p,n);
    hset->setCoordinateSystem(B);
    hset->r[0] = 0.;                 // section coordinate
    hset->r[1] = 12.*interval(-r,r); // unstable coordinate we make 12 times wider.
    for(int i=2;i<dim;++i)
      hset->r[i] = interval(-r,r)/i;
    if(dim>14)
      hset->r[14] = interval(-r,r);
    hset->iterations = period;             
    return hset;
  }

  /// read approximate heteroclinic trajectory from the file - precomputed data.
  int readHeteroclinicChain(const char* _in, std::vector<LDVector>& c, int dim){
    int i,N,j;
    LDVector u(14);

    std::ifstream in;
    in.open(_in);
    in >> N;

    for(int j=0;j<=N;++j)
    { 
      in >> u;
      c.push_back(resizeVector(u,dim));
    }
    in.close();
    return N;
  }

  /// This routine creates HSets aling heteroclinic chain between a1 and a2
  void initP1P2Chain(LDKSPoincareMap& pm, int dim){
    std::vector<LDVector> centers;
    /// first read point from the file (precomputed)
    int N = readHeteroclinicChain("data/h12.dat",centers,dim);
    h12.resize(11);

    /// Leading points in the heteroclinic sequence are very close to periodic point p1.
    /// We set the section at them to be parallel to those at p1.
    /// We also inherit coordinate system on these sections from p1.

    /// NOTE! The constants below are hand-chosen and they correspond to sizes of HSets.
    /// After simulations and many tests we have found that with these constants our routine
    /// that compute Poincare map returns bounds that are sufficient to check covering relations.
    for(int i=0;i<11;++i){
      if(i<9){
        h12[i] = new HSet(*fp1);
        h12[i]->iterations = 5;
        if(i==7) 
          h12[i]->r[1] *= 1.25;                      
        if(i==8){
          h12[i]->r[2] *= 4.;
        }
      }
      else
        h12[i] = new HSet(*fp2);      
      h12[i]->r[1] *= 0.75;
      LDVector n = pm.setOrthogonalSectionAt(centers[4*i]);
      h12[i]->setCenterAndSection(centers[4*i],n);
    }
    h12[10]->iterations = 5;
    h12[0]->r[1] /= 1.5;
    
    for(int i=2;i<14;++i){
      h12[0]->r[i] /= 4;
      h12[10]->r[i] /= 5;      
    }
  }

  /// This routine creates HSets aling heteroclinic chain between a1 and a2
  void initP2P1Chain(LDKSPoincareMap& pm, int dim){
    std::vector<LDVector> centers;    
    /// first read point from the file (precomputed)
    int N = readHeteroclinicChain("data/h21.dat",centers,dim);
    h21.resize(11);
    int i;
    LDVector n(dim);

    /// Leading points in the heteroclinic sequence are very close to periodic point p2.
    /// We set the section at them to be parallel to those at p2.
    /// We also inherit coordinate system on these sections from p2.

    /// NOTE! The constants below are hand-chosen and they correspond to sizes of HSets.
    /// After simulations and many tests we have found that with these constants our routine
    /// that compute Poincare map returns bounds that are sufficient to check covering relations.
    for(i=0;i<9;++i){
      h21[i] = new HSet(*fp2);                
      n = pm.setOrthogonalSectionAt(centers[4*i]);
      h21[i]->setCenterAndSection(centers[4*i],n);
      if(i%2==0)
        h21[i]->iterations = 5;
      else 
        h21[i]->iterations = 4;
    }

    h21[0]->r[1] /= 2;
    for(i=2;i<14;++i)
      h21[0]->r[i] /= 5;
    
    h21[8]->iterations = 6;
    h21[8]->r[1] /=2.75; 
    h21[8]->r[2] *=1.7;

    i=9;
    h21[i] = new HSet(*fp1);
    h21[i]->r[1] *= 0.6;
    h21[i]->r[2] *= 1.15;
    n = pm.setOrthogonalSectionAt(centers[4*i+2]);
    h21[i]->setCenterAndSection(centers[4*i+2],n);
    h21[i]->iterations = 5;

    h21[10] = new HSet(*fp1);
    h21[10]->r[1] /= 1.5;
    for(i=2;i<14;++i)
      h21[10]->r[i] /= 5;
    i=10;
    n = pm.setOrthogonalSectionAt(centers[4*i+2]);
    h21[i]->setCenterAndSection(centers[4*i+2],n);
    h21[i]->iterations = 3;
  }
  
  void destroy(){
    delete fp1;
    delete fp2;
    for(auto p : h12) delete p;
    for(auto p : h21) delete p;
  }
  
} data;
